// Mega 644 version

// Camera Y at Port A
// Camera datapins
// Port C.0 SCL
// Port C.1 SDA                      
// Port C.2 HREF
// Port C.3 VSYN
// Port D.2 PCLK             
#include <avr/io.h>
#include <stdlib.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <string.h>
#include <inttypes.h>
#include <stdio.h>
//I2C communication
#include "i2cmaster.h"
//UART communication
#include "uart.h"
//Rotational communication
//#include "matrix.h"

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

//Camera defines
//image width and height
//QCIF 176x144 format, shrunk window size to 120x120
#define MAXLINES 30
#define MAXPIXELS 30
#define CAMADDR 0xc0
#define SAMPLE 60
#define dataSize (MAXPIXELS/SAMPLE)

#define HREF PORTC & (1<<PINC2)  //PINC.2
#define VSYN PORTC & (1<<PINC3)  //PINC.3
#define PCLK PORTD & (1<<PIND2)  //PIND.2

#define FCPU 16000000UL

char cameraOn;
unsigned char lineBuffer[MAXLINES][MAXPIXELS];

void turnOffCamera();
void sendLineBuffer(char pixNum);

volatile unsigned char pixelNum;
volatile unsigned char lineNum;
volatile unsigned char even;
char count;
//State Machine Defines
//Camera State Machine
volatile char syncState;
#define VSYNC        1
#define HREFUP    2
#define HREFDWN   3
//Main State Machine
volatile char commandFlag;
#define CAPTURE        1
#define SEND        2
#define IDLE         0

void initialize(void); //all the usual mcu stuff

//**********************************************************
//ISR defines for camera state

ISR(PCINT2_vect)
{     
	PORTD = PORTD ^ (1<<PIND7);

	switch (syncState)
	{
	case VSYNC:
		
		syncState = HREFUP;
		//reset line number
		lineNum = 0;
		//turn on HREF ISR
		PCMSK2 = (1<<PCINT18);
		break;
		
	case HREFUP:
		pixelNum = 0;
		syncState = HREFDWN;

		//turn on pixel clock pos edge
		EIMSK |= (1<<INT0);
		even = 1;
		//pixel clock triggers on positive edge
		//EICRA |= (1<<ISC01) | (1<<ISC00); -- I put this in initalize()
		
		break;

	case HREFDWN:    
		PORTD = 0x00;

		//increment line number
		lineNum++;
		//turn off pixel counter            
		EIMSK = 0x00;

		if(lineNum >= MAXLINES)
		{
			commandFlag = SEND;
			syncState = VSYNC;
			//turn off href ISR, single image capture
			PCMSK2 = 0x00;
			//turn on VREF
			//PCMSK2 = (1<<PCINT19);
		}

		else
		{
			syncState = HREFUP;
		//	commandFlag = UPDATE;			

		}

		break;
	}

	//Clear the flag
	//PORTD = PORTD ^ (1<<PIND6);
	//PCIFR = 0x00;

}

///Pxclk (PIND2)
ISR(INT0_vect)
{         
	if (pixelNum != MAXPIXELS)
	{
		lineBuffer[lineNum][pixelNum] = PINA;
		pixelNum++;
		EIFR = 0x00;
	}
	else {
		//captured pixels from line, end pixel clock
		EIMSK = 0x00;
	}
}

//**********************************************************
//Camera I2C Communication
/// I2C read from camera
unsigned char cameraRead(unsigned char regNum)
{
	unsigned char write;
	unsigned char read;
	///Check these if they start to cause trouble
	write = i2c_start(CAMADDR+I2C_WRITE);   
	write = i2c_write(regNum);          
	i2c_stop();                              
	read = i2c_start(CAMADDR+I2C_READ); ;
	read = i2c_readNak();
	i2c_stop();                              
	return read;
}

/// I2C Write to camera
unsigned char cameraWrite(unsigned char regNum, unsigned char data) {
	unsigned char write;
	///Check these if they start to cause trouble
	write = i2c_start(CAMADDR+I2C_WRITE);   
	write = i2c_write(regNum);    
	write = i2c_write(data);          
	i2c_stop();       
	return 0;
}

///its actually already on, we are just deciding to listen to it here
void turnOnCamera()
{

	//PINC3 interrupt set on
	PCICR = (1<<PCIE2);
	PCMSK2 = (1<<PCINT19);//VREF PINC3
	cameraOn = 1;
}

///Ignore the camera's chatter
void turnOffCamera()
{
	///Turn off camera interrupts
	PCICR = 0x00;//&= ~(1<<PCIE2);
	PCMSK2 = 0x00;//&= ~(1<<PCINT19) & ~(1<<PCINT18);

	//turn off pixel counter
	EIMSK &= ~(1<<INT0);
	cameraOn = 0;
}

void initCam(){
	_delay_ms(10);
	char w1 = cameraWrite(0x12, 0x80);  // initiate a soft reset to make sure that all the regsiters are at default values
	_delay_ms(1000);
	//char w2 = cameraWrite(0x11, 0x2e);  //default is 0. slow PCLK by 96 times = 165kHz
	char w2 = cameraWrite(0x11, 0x10);  //default is 0. slow PCLK by 96 times = 165kHz
	_delay_ms(10);
	char w3 = cameraWrite(0x12, 0x24);  //COMA: default is x24. YCrCb, disable AWB, 16bit format
	_delay_ms(10);
	// char w4 = cameraWrite(0x13, 0x01);  //COMB: default is x01.
	//_delay_ms(10);
	char w5 = cameraWrite(0x14, 0x20);  //COMC: default is x00. turn on QCIF
	_delay_ms(10);
	//cameraWrite(0x28, 0x40); //enable BW mode
	_delay_ms(10);
	//Smaller window size: 176 - 120 = 56 or 28 to 148
	//144 - 120 = 24 or 12 to 132
	w2 = cameraWrite(0x17,0x56);  
	_delay_ms(10);
	w3 = cameraWrite(0x18,0xCE);  
	_delay_ms(10);
	char w4 = cameraWrite(0x19,0x0E); 
	_delay_ms(10);
	w5 = cameraWrite(0x1A,0x86);  
	_delay_ms(10);

//	fprintf(stdout, "WRITE: %d %d %d %d %d\n", w1, w2, w3, w4,  w5);

//Exposure
cameraWrite(0x10, 0x01); //2*HREF*1, shortest exposure time
_delay_ms(10);
cameraWrite(0x25, 0xca);
_delay_ms(10);
cameraWrite(0x24, 0x01);
_delay_ms(10);

///Gain control
cameraWrite(0x01, 0x00);
_delay_ms(10);
cameraWrite(0x02, 0x01);
_delay_ms(10);

//Desaturate image
cameraWrite(0x03, 0x00);
_delay_ms(10);
cameraWrite(0x4F, 0x03); //color kill storngest
_delay_ms(10);
//Adjust bias to full signal range
cameraWrite(0x34, 0x00);
_delay_ms(10);

//Max contrast, lowest brightness
cameraWrite(0x05, 0xff);
_delay_ms(10);
cameraWrite(0x06,0x00);
_delay_ms(10);

}

//**********************************************************
///Send data to PC
void sendLineBuffer(char pixNum)
{
	long loc = 0;
	//Dump data on serial port
	fprintf(stdout, "%02x", lineNum);
	while (loc < pixNum)
	{
		loc += dataSize;
	}
	fprintf(stdout, "\n");
	//count++;
	//fprintf(stdout, "L%d %d %d\n", lineBuffer[0], lineBuffer[70], lineBuffer[125]);
	//fprintf(stdout, "L%d %d %d\n", pixNum, lineNum, count);
}

void sendImage()
{
	for (int i = 0; i < MAXLINES; i++) {
		for (int j = 0; j < MAXPIXELS; j ++) {
			fprintf(stdout, "%d\n",lineBuffer[i][j]);
		}

	}
}

//**********************************************************
//Set it all up
void initialize(void)
{
	//init the UART -- uart_init() is in uart.c
	uart_init();
	stdout = stdin = stderr = &uart_str;
	
	///Set PORTA as an input
	DDRA = 0x00;
	///Set External interrupt pins as inputs
	//CHANGED -- if input low not high
	DDRD = DDRD & ~(1 << PIND2);
	DDRC = DDRC & ~(1 << PINC2) & ~(1 << PINC3);

//	DDRC = DDRC & ~(1<< PINC0);
	//DDRC = DDRC & ~(1<< PINC1);
	//PORTC = PORTC | (1<< PINC0);
	//PORTC = PORTC | (1<< PINC1);

	///Set up i2c communcation
	i2c_init();

	cameraOn = 0;

	//pixel clock triggers on positive edge
	EICRA |= (1<<ISC01) | (1<<ISC00);

}  

//**********************************************************
//Run the code
int main()
{
	initialize();
//	fprintf(stdout, "START\n");

	_delay_ms(100);
	initCam();
	_delay_ms(100);

	//crank up ISRs
	sei ();

	int test;

	commandFlag = CAPTURE;
	syncState = VSYNC;
	turnOnCamera();
	//single image capture: do not turn on camera again!
	while(1)
	{
		switch (commandFlag)
		{
		case CAPTURE:
			if (!cameraOn)
			{
				turnOnCamera();
			}
			break;
		case SEND:
			turnOffCamera();
			//Send entire image
			sendImage();
			//Wait for processor response
		//	fscanf(stdout, 'FINISH');
			commandFlag = IDLE;
			break;
		case IDLE:
			
			break;

		}

		//clear all flags
		//PCIFR = 0x00;
	}

}
//==================================================

